#!/usr/bin/python
from __future__ import division
from collections import OrderedDict

import argparse
import os
import json
# Make it work for Python 2+3 and with Unicode
import io
try:
  to_unicode = unicode
except NameError:
  to_unicode = str

cpu_freq = 0

def merge_json_files(source):
  json_object = {}
  json_object["ranks"] = []

  #get measurement files
  file_list = os.listdir(source)
  for item in file_list:
    json_rank = {}

    #determine mpi rank based on file name (rank_#)
    rank = item.split('_', 1)[1]
    rank = rank.rsplit('.', 1)[0]
    #print("rank: {}".format(rank))

    json_rank["id"] = rank

    #open meaurement file
    file_name = str(source) + "/rank_" + str(rank) 

    try:
      with open(file_name) as json_file:
        #keep order of all objects
        data = json.load(json_file, object_pairs_hook=OrderedDict)
    except IOError as ioe:
      print("Cannot open file {} ({})".format(file_name, repr(ioe)))
      return

    #determine cpu frequency
    global cpu_freq
    cpu_freq = int(data['cpu in mhz']) * 1000000

    #get all threads
    json_rank["threads"] = data["threads"]

    #append current rank to json file
    json_object["ranks"].append(json_rank)

  # print json.dumps(json_object,indent=2, sort_keys=False,
  #                  separators=(',', ': '), ensure_ascii=False)
  return json_object

class Sum_Counters(object):
  regions = OrderedDict()
  regions_last_rank_id = {}

  def add_region(self, rank_id, region, events=OrderedDict()):

    #clean events from read values caused by PAPI_hl_read
    cleaned_events = OrderedDict()
    for key,value in events.items():
      metric_value = value
      if isinstance(value, dict):
        metric_value = float(value['total'])
      cleaned_events[key] = metric_value

    if region not in self.regions:
      #new region
      new_region_events = cleaned_events.copy()
      new_region_events['Number of ranks'] = 1
      new_region_events['Number of threads'] = 1
      new_region_events['Number of processes'] = 1
      self.regions[region] = new_region_events.copy()
      self.regions_last_rank_id[region] = rank_id
    else:
      #add counter values to existing region
      known_events = self.regions[region].copy()
      new_events = cleaned_events.copy()

      #increase number of ranks when rank_id has changed
      if self.regions_last_rank_id[region] == rank_id:
        new_events['Number of ranks'] = 0
      else:
        self.regions_last_rank_id[region] = rank_id
        new_events['Number of ranks'] = 1

      #always increase number of threads
      new_events['Number of threads'] = 1
      new_events['Number of processes'] = 1

      #add values
      for event_key,event_value in known_events.items():
        if 'Number of' in event_key or 'count' in event_key:
          known_events[event_key] = event_value + new_events[event_key]
        else:
          known_events[event_key] = float(format(event_value + new_events[event_key], '.2f'))
      self.regions[region] = known_events.copy()

  def get_json(self):
    #calculate correct thread number (number of processes / number of ranks)
    for name in self.regions:
      events = self.regions[name]
      events['Number of threads'] = int(events['Number of processes'] / events['Number of ranks'])
    return self.regions


def sum_json_object(json):
  sum_cnt = Sum_Counters()
  for ranks in json['ranks']:
    for threads in ranks['threads']:
      for regions in threads['regions']:
        for region_key,region_value in regions.items():
          name = region_key
          events = region_value
          sum_cnt.add_region(ranks['id'], name, events)

  return sum_cnt.get_json()


def get_ipc_dict(inst, cyc):
  ipc_dict = OrderedDict()
  for (inst_key,inst_value), (cyc_key,cyc_value) in zip(inst.items(), cyc.items()):
    #print str(inst_key) + "," + str(inst_value)
    #print str(cyc_key) + "," + str(cyc_value)
    try:
      ipc = float(int(inst_value) / int(cyc_value))
    except:
      ipc = 0
    ipc_dict[inst_key] = float(format(ipc, '.2f'))
  return ipc_dict


def get_ops_dict(ops, rt):
  ops_dict = OrderedDict()
  for (ops_key,ops_value), (rt_key,rt_value) in zip(ops.items(), rt.items()):
    #print str(ops_key) + "," + str(ops_value)
    #print str(rt_key) + "," + str(rt_value)
    ops = float(ops_value) / 1000000 / rt_value
    ops_dict[ops_key] = float(format(ops, '.2f'))
  return ops_dict


def convert_value(value, event_type = 'Other'):
  if event_type == 'Other':
    result = float(value)
    result = float(format(result, '.2f'))
  elif event_type == 'Runtime':
    try:
      result = float(value) / int(cpu_freq)
    except:
      result = 1.0
    result = float(format(result, '.2f'))
  elif event_type == 'CPUtime':
    result = float(value) / 1000000000
    result = float(format(result, '.2f'))

  return result


def format_read_events(events, event_type = 'Other'):
  format_read_dict = OrderedDict()
  for read_key,read_value in events.items():
    format_read_dict[read_key] = convert_value(read_value, event_type)
  return format_read_dict


def format_events(events):
  #keep order as declared
  format_events = OrderedDict()
  #remember runtime for other metrics like MFLOPS
  rt = 1.0
  rt_dict = OrderedDict()

  #Region Count
  if 'region_count' in events:
    format_events['Region count'] = int(events['region_count'])
    del events['region_count']

  #Real Time
  if 'cycles' in events:
    if isinstance(events['cycles'],dict):
      for read_key,read_value in events['cycles'].items():
        rt_dict[read_key] = float(read_value) / int(cpu_freq)
      format_events['Real time in s'] = format_read_events(events['cycles'],'Runtime')
    else:
      rt = float(events['cycles']) / int(cpu_freq)
      format_events['Real time in s'] = convert_value(events['cycles'], 'Runtime')
    del events['cycles']

  #CPU Time
  if 'perf::TASK-CLOCK' in events:
    if isinstance(events['perf::TASK-CLOCK'],dict):
      format_events['CPU time in s'] = format_read_events(events['perf::TASK-CLOCK'],'CPUtime')
    else:
      format_events['CPU time in s'] = convert_value(events['perf::TASK-CLOCK'], 'CPUtime')
    del events['perf::TASK-CLOCK']

  #PAPI_TOT_INS and PAPI_TOT_CYC to calculate IPC
  if 'PAPI_TOT_INS' and 'PAPI_TOT_CYC' in events:
    if isinstance(events['PAPI_TOT_INS'],dict) and isinstance(events['PAPI_TOT_CYC'],dict):
      ipc_dict = get_ipc_dict(events['PAPI_TOT_INS'], events['PAPI_TOT_CYC'])
      format_events['IPC'] = ipc_dict
    else:
      try:
        ipc = float(int(events['PAPI_TOT_INS']) / int(events['PAPI_TOT_CYC']))
      except:
        ipc = 0
      format_events['IPC'] = float(format(ipc, '.2f'))

    del events['PAPI_TOT_INS']
    del events['PAPI_TOT_CYC']
  
  #FLIPS
  if 'PAPI_FP_INS' in events:
    if isinstance(events['PAPI_FP_INS'],dict):
      mflips_dict = get_ops_dict(events['PAPI_FP_INS'], rt_dict)
      format_events['MFLIPS/s'] = mflips_dict
    else:
      mflips = float(events['PAPI_FP_INS']) / 1000000 / rt
      mflips = float(format(mflips, '.2f'))
      format_events['MFLIPS/s'] = mflips
    del events['PAPI_FP_INS']

  #SP vector instructions per second
  if 'PAPI_VEC_SP' in events:
    if isinstance(events['PAPI_VEC_SP'],dict):
      mvecins_dict = get_ops_dict(events['PAPI_VEC_SP'], rt_dict)
      format_events['Single precision vector/SIMD instructions rate in M/s'] = mvecins_dict
    else:
      mvecins = float(events['PAPI_VEC_SP']) / 1000000 / rt
      mvecins = float(format(mvecins, '.2f'))
      format_events['Single precision vector/SIMD instructions rate in M/s'] = mvecins
    del events['PAPI_VEC_SP']

  #DP vector instructions per second
  if 'PAPI_VEC_DP' in events:
    if isinstance(events['PAPI_VEC_DP'],dict):
      mvecins_dict = get_ops_dict(events['PAPI_VEC_DP'], rt_dict)
      format_events['Double precision vector/SIMD instructions rate in M/s'] = mvecins_dict
    else:
      mvecins = float(events['PAPI_VEC_DP']) / 1000000 / rt
      mvecins = float(format(mvecins, '.2f'))
      format_events['Double precision vector/SIMD instructions rate in M/s'] = mvecins
    del events['PAPI_VEC_DP']
  
  #FLOPS
  if 'PAPI_FP_OPS' in events:
    if isinstance(events['PAPI_FP_OPS'],dict):
      mflops_dict = get_ops_dict(events['PAPI_FP_OPS'], rt_dict)
      format_events['MFLOPS/s'] = mflops_dict
    else:
      mflops = float(events['PAPI_FP_OPS']) / 1000000 / rt
      mflops = float(format(mflops, '.2f'))
      format_events['MFLOPS/s'] = mflops
    del events['PAPI_FP_OPS']
  
  #SP FLOPS
  if 'PAPI_SP_OPS' in events:
    if isinstance(events['PAPI_SP_OPS'],dict):
      mflops_dict = get_ops_dict(events['PAPI_SP_OPS'], rt_dict)
      format_events['Single precision MFLOPS/s'] = mflops_dict
    else:
      mflops = float(events['PAPI_SP_OPS']) / 1000000 / rt
      mflops = float(format(mflops, '.2f'))
      format_events['Single precision MFLOPS/s'] = mflops
    del events['PAPI_SP_OPS']

  #DP FLOPS
  if 'PAPI_DP_OPS' in events:
    if isinstance(events['PAPI_DP_OPS'],dict):
      mflops_dict = get_ops_dict(events['PAPI_DP_OPS'], rt_dict)
      format_events['Double precision MFLOPS/s'] = mflops_dict
    else:
      mflops = float(events['PAPI_DP_OPS']) / 1000000 / rt
      mflops = float(format(mflops, '.2f'))
      format_events['Double precision MFLOPS/s'] = mflops
    del events['PAPI_DP_OPS']

  #read the rest
  for event_key,event_value in events.items():
    if isinstance(event_value,dict):
      format_events[event_key] = format_read_events(event_value)
    else:
      format_events[event_key] = convert_value(event_value)

  return format_events


def format_json_object(json):
  json_object = {}
  json_object['ranks'] = []

  for rank in json['ranks']:
    # print rank['id']
    # print rank['threads']
    json_rank = {}
    json_rank['id'] = rank['id']
    json_rank['threads'] = []

    for thread in rank['threads']:
      # print thread['id']
      json_thread = {}
      json_thread['id'] = thread['id']
      json_thread['regions'] = []
      for region in thread['regions']:
        json_region = {}
        for region_key,region_value in region.items():
          # print region_key
          # print region_value
          json_region[region_key] = format_events(region_value)

        json_thread['regions'].append(json_region)
      json_rank['threads'].append(json_thread)
    json_object['ranks'].append(json_rank)

  return json_object

def write_json_file(data, file_name):
  with io.open(file_name, 'w', encoding='utf8') as outfile:
    str_ = json.dumps(data,
                      indent=4, sort_keys=False,
                      separators=(',', ': '), ensure_ascii=False)
    outfile.write(to_unicode(str_))
    print (str_)


def main(source, format, type):
  if (format == "json"):
    json = merge_json_files(source)
    formated_json = format_json_object(json)

    if type == 'detail':
      write_json_file(formated_json, 'papi.json')

    #summarize data over threads and ranks
    if type == 'accumulate':
      sum_json = sum_json_object(formated_json)
      write_json_file(sum_json, 'papi_sum.json')
  else:
    print("Format not supported!")


def parse_args():
  parser = argparse.ArgumentParser()
  parser.add_argument('--source', type=str, required=False, default="papi_hl_output",
                      help='Measurement directory of raw data.')
  parser.add_argument('--format', type=str, required=False, default='json', 
                      help='Output format, e.g. json.')
  parser.add_argument('--type', type=str, required=False, default='detail', 
                      help='Output type: detail or accumulate.')

  # check if papi directory exists
  source = str(parser.parse_args().source)
  if os.path.isdir(source) == False:
    print("Measurement directory '{}' does not exist!\n".format(source))
    parser.print_help()
    parser.exit()

  # check format
  output_format = str(parser.parse_args().format)
  if output_format != "json":
    print("Output format '{}' is not supported!\n".format(output_format))
    parser.print_help()
    parser.exit()

  # check type
  output_type = str(parser.parse_args().type)
  if output_type != "detail" and output_type != "accumulate":
    print("Output type '{}' is not supported!\n".format(output_type))
    parser.print_help()
    parser.exit()
  

  return parser.parse_args()


if __name__ == '__main__':
  args = parse_args()
  main(format=args.format,
       source=args.source,
       type=args.type)


#
# Copyright (c) 2008 Google, Inc.
# Contributed by Arun Sharma <arun.sharma@google.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import print_function
import os
from perfmon import *

def public_members(self):
    s = "{ "
    for k, v in self.__dict__.items():
      if not k[0] == '_':
        s += "%s : %s, " % (k, v)
    s += " }"
    return s

class System:
  # Use the os that gives us everything
  os = PFM_OS_PERF_EVENT_EXT

  def __init__(self):
    self.ncpus = os.sysconf('SC_NPROCESSORS_ONLN')
    self.pmus = []
    for i in range(0, PFM_PMU_MAX):
      try:
        pmu = PMU(i)
      except:
        pass
      else:
        self.pmus.append(pmu)

  def __repr__(self):
    return public_members(self)

class Event:
  def __init__(self, info):
    self.info = info
    self.__attrs = []

  def __repr__(self):
    return '\n' + public_members(self)

  def __parse_attrs(self):
    info = self.info
    for index in range(0, info.nattrs):
      self.__attrs.append(pfm_get_event_attr_info(info.idx, index,
                                                  System.os)[1])

  def attrs(self):
    if not self.__attrs:
      self.__parse_attrs()
    return self.__attrs

class PMU:
  def __init__(self, i):
    self.info = pfm_get_pmu_info(i)[1]
    self.__events = []

  def __parse_events(self):
    index = self.info.first_event
    while index != -1:
      self.__events.append(Event(pfm_get_event_info(index, System.os)[1]))
      index = pfm_get_event_next(index)

  def events(self):
    if not self.__events:
      self.__parse_events()
    return self.__events

  def __repr__(self):
    return public_members(self)

if __name__ == '__main__':
  from perfmon import *
  s = System()
  for pmu in s.pmus:
    info = pmu.info
    if info.flags.is_present:
      print(info.name, info.size, info.nevents)
      for e in pmu.events():
        print(e.info.name, e.info.code)
        for a in e.attrs():
          print('\t\t', a.name, a.code)



#!/usr/bin/env python

from distutils.core import setup, Extension
from distutils.command.install_data import install_data

setup(name='perfmon',
      version='0.1',
      author='Arun Sharma',
      author_email='arun.sharma@google.com',
      description='libpfm wrapper',
      packages=['perfmon'],
      package_dir={ 'perfmon' : 'src' },
      py_modules=['perfmon.perfmon_int'],
      ext_modules=[Extension('perfmon._perfmon_int',
                  sources = ['src/perfmon_int.i'],
                  libraries = ['pfm'],
                  library_dirs = ['../lib'],
                  include_dirs = ['../include'],
                  swig_opts=['-I../include'])])



#!/usr/bin/env python
#
# Copyright (c) 2008 Google, Inc.
# Contributed by Arun Sharma <arun.sharma@google.com>
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 
# Self monitoring example. Copied from self.c

from __future__ import print_function
import os
import optparse
import random
import errno
import struct
import perfmon

if __name__ == '__main__':
  parser = optparse.OptionParser()
  parser.add_option("-e", "--events", help="Events to use",
                    action="store", dest="events")
  parser.set_defaults(events="PERF_COUNT_HW_CPU_CYCLES")
  (options, args) = parser.parse_args()

  if options.events:
    events = options.events.split(",")
  else:
    raise Exception("You need to specify events to monitor")

  s = perfmon.PerThreadSession(int(os.getpid()), events)
  s.start()

  # code to be measured
  #
  # note that this is not identical to what examples/self.c does
  # thus counts will be different in the end
  for i in range(1, 1000000):
    random.random()

  # read the counts
  for i in range(0, len(events)):
    count = struct.unpack("L", s.read(i))[0]
    print("""%s\t%lu""" % (events[i], count))



#!/usr/bin/env python
#
# Copyright (c) 2008 Google, Inc.
# Contributed by Arun Sharma <arun.sharma@google.com>
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# 

import os
from perfmon import *

def public_members(self):
    s = "{ "
    for k, v in self.__dict__.iteritems():
      if not k[0] == '_':
        s += "%s : %s, " % (k, v)
    s += " }"
    return s

class System:
  def __init__(self):
    self.ncpus = os.sysconf('SC_NPROCESSORS_ONLN')
    self.pmu = PMU()

  def __repr__(self):
    return public_members(self)

class Event:
  def __init__(self):
    pass

  def __repr__(self):
    return '\n' + public_members(self)

class EventMask:
  def __init__(self):
    pass

  def __repr__(self):
    return '\n\t' + public_members(self)

class PMU:
  def __init__(self):
    pfm_initialize()

    self.type = pfm_py_get_pmu_type()
    self.name = pfm_get_pmu_name(PFMON_MAX_EVTNAME_LEN)[1]
    self.width = pfm_py_get_hw_counter_width()

    # What does the PMU support?
    self.__implemented_pmcs = pfmlib_regmask_t()
    self.__implemented_pmds = pfmlib_regmask_t()
    self.__implemented_counters = pfmlib_regmask_t()
    pfm_get_impl_pmcs(self.__implemented_pmcs)
    pfm_get_impl_pmds(self.__implemented_pmds)
    pfm_get_impl_counters(self.__implemented_counters)
    self.implemented_pmcs = self.__implemented_pmcs.weight()
    self.implemented_pmds = self.__implemented_pmds.weight()
    self.implemented_counters = self.__implemented_counters.weight()

    self.__events = None

  def __parse_events(self):
    nevents = pfm_py_get_num_events()
    self.__events = []
    for idx in range(0, nevents):
      e = Event()
      e.name = pfm_py_get_event_name(idx)
      e.code = pfm_py_get_event_code(idx)
      e.__counters = pfmlib_regmask_t()
      pfm_get_event_counters(idx, e.__counters)
      # Now the event masks
      e.masks = []
      nmasks = pfm_py_get_num_event_masks(idx)
      for mask_idx in range(0, nmasks):
        em = EventMask()
        em.name = pfm_py_get_event_mask_name(idx, mask_idx)
        em.code = pfm_py_get_event_mask_code(idx, mask_idx)
        em.desc = pfm_py_get_event_mask_description(idx, mask_idx)
        e.masks.append(em)
      self.__events.append(e)

  def events(self):
    if not self.__events:
      self.__parse_events()
    return self.__events

  def __repr__(self):
    return public_members(self)

if __name__ == '__main__':
  from perfmon import *
  s = System()
  print s
  print s.pmu.events()
